from pyshrimp.execution_pipeline.pipeline import ExecutionPipeline
from pyshrimp.execution_pipeline.pipeline_api import PipelineTerminator, PipelineTerminatorStdout


# noinspection PyMethodMayBeStatic
class PipelineStarter:

    def empty(self):
        return ExecutionPipeline()

    def stdin(self):
        return self.empty().attach_stdin()

    def text(self, text: str):
        return self.empty().attach_text(text)

    def close(self) -> PipelineTerminator:
        return PipelineTerminator()

    def __or__(self, other) -> ExecutionPipeline:
        return self.empty() | other

    def __ror__(self, other):
        return self.empty() | other


PIPE = PipelineStarter()
PIPE_END = PipelineTerminator()
PIPE_END_STDOUT = PipelineTerminatorStdout()



from contextlib import contextmanager
from typing import Type

from pyshrimp._internal.locking.file_based_lock_internal import FileBasedLockInternal
from pyshrimp._internal.utils.platformspecific import running_on_windows


def _get_file_based_lock_impl() -> Type[FileBasedLockInternal]:
    if running_on_windows():
        from pyshrimp._internal.locking.file_based_lock_windows import get_file_based_lock_impl_for_windows
        return get_file_based_lock_impl_for_windows()
    else:
        from pyshrimp._internal.locking.file_based_lock_posix import get_file_based_lock_impl_for_posix
        return get_file_based_lock_impl_for_posix()


_FileBasedLockImpl = _get_file_based_lock_impl()

class FileBasedLock:

    def __init__(self, lock_file_path):
        self._delegate: FileBasedLockInternal = _FileBasedLockImpl(
            lock_file_path=lock_file_path
        )

    def try_lock(self):
        return self._delegate.try_lock()

    def release_lock(self):
        return self._delegate.release_lock()

    @contextmanager
    def acquire_lock(self):
        locked = self.try_lock()
        try:
            yield locked
        finally:
            if locked:
                self.release_lock()


@contextmanager
def acquire_file_lock(lock_file_path):
    with FileBasedLock(lock_file_path).acquire_lock() as locked:
        yield locked




import subprocess
from typing import Union, Iterable

from pyshrimp.utils.string_wrapper import StringWrapper

RETURN_CODE_EXECUTION_FAILED = 300


class ProcessExecutionResult(object):
    def __init__(
            self,
            command: Union[str, Iterable],
            standard_output: StringWrapper,
            error_output: StringWrapper,
            return_code: int,
            exception=None
    ):
        self.command = command
        self.standard_output = standard_output
        self.error_output = error_output
        self.return_code = return_code
        self.exception = exception

    def is_ok(self):
        return self.return_code == 0

    def raise_if_not_ok(self):
        if not self.is_ok():
            reason = (
                f'Got exception: {self.exception}'
                if self.return_code == RETURN_CODE_EXECUTION_FAILED
                else f'Command returned non zero exit code: {self.return_code}'
            )

            raise ProcessExecutionException(
                message=f'Command execution failed: {reason}',
                result=self
            )
        return self

    def execution_report(self, message: str):
        command_as_string = self.command if isinstance(self.command, str) else ' '.join(self.command)
        report = (
            f'{message}\n'
            f'  Command: {command_as_string}\n'
            f'  Exit code: {self.return_code}\n'
            f'  Stdout: {self.standard_output!r}\n'
            f'  Error: {self.error_output!r}'
        )
        if self.exception:
            report += f'\n  Exception: {self.exception}'

        return report

    def __str__(self, *args, **kwargs):
        return 'CommandExecutionResult{}'.format(repr(self.__dict__))

    @property
    def out(self):
        return self.standard_output

    @property
    def err(self):
        return self.error_output

    @property
    def exit_code(self):
        return self.return_code


class ProcessExecutionException(Exception):

    def __init__(self, message: str, result: ProcessExecutionResult) -> None:
        super().__init__(message)
        self.result = result
        self.message = message

    def __str__(self):
        return self.result.execution_report(self.message)


def run_process(command: Union[str, Iterable], timeout=None, cmd_in=None, capture_out=True, capture_err=True,
                run_in_shell=False, cwd=None, env=None) -> ProcessExecutionResult:
    try:
        p = subprocess.Popen(
            command,
            stdin=None if cmd_in is None else subprocess.PIPE,
            stdout=subprocess.PIPE if capture_out else None,
            stderr=subprocess.PIPE if capture_err else None,
            universal_newlines=True,
            shell=run_in_shell,
            cwd=cwd,
            env=env
        )

        (out, err) = p.communicate(input=cmd_in, timeout=timeout)

        return ProcessExecutionResult(
            command=command,
            standard_output=StringWrapper(out or ''),
            error_output=StringWrapper(err or ''),
            return_code=p.returncode
        )

    except Exception as e:
        return ProcessExecutionResult(
            command=command,
            standard_output=StringWrapper(''),
            error_output=StringWrapper(str(e)),
            return_code=RETURN_CODE_EXECUTION_FAILED,
            exception=e
        )



import re


def re_match_all(elements, pattern, capture_group=1):
    matches = [
        re.match(pattern, el) for el in elements
    ]

    return [
        m.group(capture_group) for m in matches if m
    ]



__version__ = '0.6.0'

# noinspection PyProtectedMember
from pyshrimp._internal.wrapper.mainwrapper import _run as run
# noinspection PyProtectedMember
from pyshrimp._internal.wrapper.mainwrapper import _init_logging as init_logging
from pyshrimp.execution_pipeline.pipeline import pipe, ExecutionPipeline
from pyshrimp.execution_pipeline.pipeline_api import PipelineExecutionResult
from pyshrimp.execution_pipeline.pipeline_starter import PIPE, PIPE_END, PIPE_END_STDOUT
from pyshrimp.utils.command import cmd, shell_cmd, Command, SkipConfig, CommandArgProcessor, DefaultCommandArgProcessor
from pyshrimp.utils.dotdict import as_dot_dict, unwrap_dot_dict
from pyshrimp.utils.filesystem import ls, glob_ls, chmod_set, chmod_unset, read_file, read_file_bin, write_to_file
from pyshrimp.utils.splitter import regex_splitter, create_regex_splitter
from pyshrimp.utils.table_parser import parse_table, ParsedTable
from pyshrimp.utils.locking import acquire_file_lock, FileBasedLock
from pyshrimp.utils.logging import log, exit_error
from pyshrimp.utils.matching import re_match_all
from pyshrimp.utils.parallel import in_background
from pyshrimp.utils.string_wrapper import StringWrapper
from pyshrimp.utils.subprocess_utils import run_process, ProcessExecutionException, ProcessExecutionResult
from pyshrimp.utils.wait import wait_until, wait_until_gen
from pyshrimp.utils.collections import first_not_null


import dataclasses
import datetime
import hashlib
import json
import os
import re
import signal
import subprocess
import sys
import traceback

import pkg_resources

from pyshrimp._internal.scriptrunner.cli import _handle_cli_maybe
from pyshrimp._internal.utils.platformspecific import running_on_windows
from pyshrimp.utils.locking import acquire_file_lock


def _running_in_dev_mode():
    import pyshrimp._internal as pyshrimp_internal
    # the internal path will be in .../src/pyshrimp/_internal/__init__.py when running in dev mode
    pyshrimp_location = os.path.dirname(pyshrimp_internal.__file__).split(os.sep)[-3]
    return pyshrimp_location == 'src'


@dataclasses.dataclass
class _EnvConfig:
    log_enabled: bool
    cache_dir: str


def _default_env_config():
    _TRUE_VALUES = ['1', 'true', 'yes']
    return _EnvConfig(
        log_enabled=os.environ.get('PYSHRIMP_LOG', '0').lower() in _TRUE_VALUES,
        cache_dir=os.path.abspath(os.environ.get('PYSHRIMP_CACHE_DIR', None) or os.path.expanduser('~/.cache/pyshrimp'))
    )

def _iter_file_lines(file_path: str):
    with open(file_path, 'r') as f:
        for line in f:
            yield line

def _resolve_path(file_path: str, relative_to: str):
    if os.path.isabs(file_path):
        return file_path

    if file_path.startswith('~'):
        return os.path.expanduser(file_path)

    return os.path.normpath(os.path.join(relative_to, file_path))
class _ScriptConfig:

    def __init__(self):
        self.wrapper_opts_raw = []
        self.wrapper_opts = {}
        self.requirements = []


# noinspection PyMethodMayBeStatic
class _ScriptRunnerBootstrap:

    def __init__(self, env_config: _EnvConfig = None):
        env_config = env_config or _default_env_config()
        self._log_enabled = env_config.log_enabled
        self._cache_dir = env_config.cache_dir
        self._venvs_dir = os.path.join(self._cache_dir, 'virtual_envs')

    # -- Logging etc

    def log(self, message: str, level: str = 'INFO'):
        if self._log_enabled or level == 'ERROR':
            print(f'[PyShrimp:bootstrap] {level}: ' + message, file=sys.stderr)

    def log_error(self, message: str):
        self.log(message, 'ERROR')

    def exit_error(self, message: str, exit_code=1, exc_info=False):
        if exc_info:
            traceback.print_exc()

        self.log_error(message)

        raise sys.exit(exit_code)

    # -- script config parsing

    def parse_key_value_opts(self, line: str, pattern: str):
        result = {}
        opts_list = self.parse_list_opts(line, pattern)
        for el in opts_list:
            kv = el.split('=', maxsplit=1)
            if len(kv) == 1:
                result[kv[0]] = 'true'
            else:
                result[kv[0]] = kv[1]

        return result

    def parse_list_opts(self, line: str, pattern: str):
        m = re.match(pattern, line.strip())
        if not m:
            raise self.exit_error(f'Failed to parse config line: {line.strip()!r}')

        parsed = [el.strip() for el in m.group(1).split(',')]
        return [el for el in parsed if el]

    def parse_file_path_opt(self, target_script_path: str, line: str, pattern: str, must_exists: bool):
        m = re.match(pattern, line.strip())
        if not m:
            raise self.exit_error(f'Failed to parse config line: {line.strip()!r}')

        script_dir = os.path.dirname(os.path.abspath(target_script_path))
        file_path = _resolve_path(
            file_path=m.group(1).strip(),
            relative_to=script_dir
        )
        
        if must_exists:
            if not os.path.exists(file_path):
                raise self.exit_error(f'Failed to parse config line - file "{file_path}" does not exists: {line.strip()!r}')

        return file_path

    def read_script_config(self, target_script) -> _ScriptConfig:
        config = _ScriptConfig()

        if _running_in_dev_mode():
            config.requirements += [
                # for local testing use the dev install approach
                '-e', os.path.abspath(
                    os.path.join(
                        os.path.dirname(__file__), os.pardir, os.pardir, os.pardir, os.pardir
                    )
                )
            ]
        else:
            config.requirements += [
                # in dist mode install version from pip, not older than the current one
                f'pyshrimp>={pkg_resources.get_distribution("pyshrimp").version}'
            ]

        with open(target_script, 'r') as f:
            for line in f:
                if not line.startswith('#'):
                    break

                # TODO: support for selecting base python (# $python: path) or/and maybe pyenv support...

                if line.startswith('# $opts:'):
                    config.wrapper_opts.update(
                        self.parse_key_value_opts(line, r'# \$opts:(.*)')
                    )
                    config.wrapper_opts_raw.append(line)

                elif line.startswith('# $requires:'):
                    config.requirements += self.parse_list_opts(line, r'# \$requires:(.*)')

                elif line.startswith('# $requirements_file:'):
                    config.requirements += [l.strip() for l in _iter_file_lines(
                        self.parse_file_path_opt(target_script, line, '# \$requirements_file:(.*)', must_exists=True)    
                    )]

        return config

    def validate_config(self, config: _ScriptConfig):
        valid = True
        if config.wrapper_opts.get('magic') != 'true':
            # magic is not active
            if config.wrapper_opts.get('devloop') == 'true':
                valid = False
                self.log_error('Detected invalid configuration - devloop is enabled but magic is not, this will not work as expected.')
                self.log_error('Possible solutions')
                self.log_error(' * remove devloop from $opts if you do not want the devloop')
                self.log_error(' * remove devloop from $opts and use run(main, devloop=True) instead')
                self.log_error(' * add magic to $opts - the power of magic will be with you')

            if config.wrapper_opts.get('elevate') == 'true':
                valid = False
                self.log_error('Detected invalid configuration - elevate is enabled but magic is not, this will not work as expected.')
                self.log_error('Possible solutions')
                self.log_error(' * remove elevate from $opts if you do not want the elevate option')
                self.log_error(' * remove elevate from $opts and use run(main, elevate=True) instead')
                self.log_error(' * add magic to $opts - the power of magic will be with you')

        if not valid:
            self.log_error('Script opts lines:')
            for line in config.wrapper_opts_raw:
                self.log_error(f' > {line.rstrip()}')

        return valid

    # -- virtual environment

    def can_use_venv(self):
        # noinspection PyBroadException
        try:
            import venv
            import ensurepip
            return True
        except Exception:
            return False

    def can_use_virtualenv(self):
        # noinspection PyBroadException
        try:
            # noinspection PyUnresolvedReferences
            import virtualenv
            return True
        except Exception:
            return False

    def prepare_virtualenv(self, config: _ScriptConfig):
        requirements = config.requirements
        requirements_hash = hashlib.sha1(':'.join(requirements).encode('utf-8')).hexdigest()
        self.log(f'Using requirements {requirements_hash}: {requirements}')
        venv_dir = os.path.join(self._venvs_dir, requirements_hash)
        setup_complete_marker_file = os.path.join(venv_dir, '.setup_completed')
        if not os.path.exists(setup_complete_marker_file):
            # create directory first so we can create lock file inside
            if not os.path.exists(venv_dir):
                os.makedirs(
                    venv_dir,
                    exist_ok=True  # it may exists due to race condition
                )

            # create lock outside of virtualenv directory as we will clear it
            setup_lock_file = os.path.join(self._venvs_dir, f'.setup_lock_{requirements_hash}')
            # TODO: consider some timed re-try logic in case of lock failure...
            with acquire_file_lock(setup_lock_file) as locked:
                if not locked:
                    raise self.exit_error(
                        f'Failed to setup virtualenv!'
                        f' Could not obtain lock under {setup_lock_file}.'
                        f' Are two scripts running the environment setup?',
                        exc_info=True
                    )

                self.log(f'Setting up venv: {venv_dir}')

                if self.can_use_venv():
                    self.setup_virtual_env_using_venv(venv_dir)

                elif self.can_use_virtualenv():
                    self.setup_virtual_env_using_virtualenv(venv_dir)

                else:
                    raise self.exit_error('Failed to setup virtual environment - cannot use venv or virtualenv!')

                pip_executable_path_candidates = [
                    os.path.join(venv_dir, 'bin', 'pip'),
                    # windows
                    os.path.join(venv_dir, 'scripts', 'pip.exe')
                ]

                pip_executable_path = next(filter(os.path.exists, pip_executable_path_candidates), None)
                if pip_executable_path is None:
                    raise self.exit_error(
                        f'Setup failed - could not locate pip executable among the candidates: {pip_executable_path_candidates}'
                    )

                if any((req.startswith('-e ') or req == '-e') for req in requirements):
                    # editable mode requested, we need to ensure that pip is fresh enough to support this
                    # solution to: '''A "pyproject.toml" file was found, but editable mode currently requires a setup.py based build.'''
                    self.log(f'Ensuring PIP is in fresh enough version to support editable requirements...')
                    pip_upgrade_result = subprocess.run(
                        [
                            pip_executable_path, 'install', 'pip>=21.2.4'
                        ], stdout=subprocess.PIPE, stderr=subprocess.PIPE
                    )
                    if pip_upgrade_result.returncode != 0:
                        raise self.exit_error(
                            f'Failed to upgrade pip!\n'
                            f'Upgrade output:\n'
                            f'{pip_upgrade_result.stdout}\n'
                            f'Error Output:\n'
                            f'{pip_upgrade_result.stderr}',
                            exc_info=True
                        )

                install_requirements_command = [pip_executable_path, 'install', '--disable-pip-version-check'] + requirements
                self.log(f'Installing requirements: {install_requirements_command}')
                pip_install_result = subprocess.run(install_requirements_command, stdout=subprocess.PIPE)

                if pip_install_result.returncode != 0:
                    raise self.exit_error(
                        f'Failed to install requirements in virtualenv! Installation output:\n'
                        f'{pip_install_result.stdout}',
                        exc_info=True
                    )

                with open(setup_complete_marker_file, 'w') as f:
                    f.write(f'Completed at: {datetime.datetime.now().isoformat()}')

                self.log(f'Virtual env ready')

        python_executable_path_candidates = [
            os.path.join(venv_dir, 'bin', 'python'),
            # windows
            os.path.join(venv_dir, 'scripts', 'python.exe')
        ]

        python_executable_path = next(filter(os.path.exists, python_executable_path_candidates), None)
        if python_executable_path is None:
            raise self.exit_error(
                f'Setup failed - could not locate python executable among the candidates: {python_executable_path_candidates}'
            )
        
        return python_executable_path

    def setup_virtual_env_using_venv(self, venv_dir):
        import venv
        try:
            self.log(f'Setting up venv using venv')
            venv.create(venv_dir, system_site_packages=True, with_pip=True, clear=True)
        except BaseException:
            raise self.exit_error(f'Failed to setup virtualenv! Check previous messages for details.', exc_info=True)

    def setup_virtual_env_using_virtualenv(self, venv_dir):
        command = [sys.executable, '-m', 'virtualenv', '-p', sys.executable, '--clear', venv_dir]
        self.log(f'Setting up venv using virtualenv: {command}')
        result = subprocess.run(command, stdout=subprocess.PIPE)
        if result.returncode != 0:
            raise self.exit_error(
                f'Failed to setup virtualenv!\n'
                f'Setup command: {command}'
                f'Setup output: {result.stdout}'
            )

    def execute_bootstrap(self):
        argv = sys.argv
        if len(argv) < 2:
            self.exit_error(f'Not enough arguments provided: {argv}')

        target_script = argv[1]
        args = argv[2:]

        self.log(f'target: {target_script}')
        self.log(f'args: {args}')

        if not os.path.exists(target_script) or not os.path.isfile(target_script):
            self.exit_error(
                f'Target script does not exist: {target_script}\n'
                f'Usage: pyshrimp target-script.py arg1 arg2 ... argN\n'
                f'For CLI options run: pyshrimp --help'
            )

        config = self.read_script_config(target_script)
        if not self.validate_config(config):
            raise self.exit_error('Script configuration is invalid, please see previous messages.')

        python_executable = self.prepare_virtualenv(config)

        if config.wrapper_opts.get('magic') == 'true':
            os.environ['PYSHRIMP_MAGIC_WRAPPER_OPTS'] = json.dumps(config.wrapper_opts)
            exec_args = [python_executable, '-u', '-m', 'pyshrimp._internal.wrapper.magicwrapper', '--', target_script] + args
        else:
            exec_args = [python_executable, '-u', target_script] + args

        if running_on_windows():
            self._exec_windows_workaround(exec_args)

        else:
            self.log(f'Executing the script: {exec_args}')
            os.execlp(exec_args[0], *exec_args)


    def _exec_windows_workaround(self, exec_args):
        self.log(f'Executing the script with subprocess: {exec_args}')

        def _on_sig_int(signum, frame):
            self.log('Got SIGINT, ignoring (should be also received by child process)')

        signal.signal(signal.SIGINT, _on_sig_int)

        proc = subprocess.Popen(exec_args)
        proc.wait()
        
        self.log(f'Process completed, exit code: {proc.returncode}')

        raise sys.exit(proc.returncode)


def _bootstrap():
    if _handle_cli_maybe():
        return
    else:
        _ScriptRunnerBootstrap().execute_bootstrap()



import os
from typing import Type

from pyshrimp._internal.locking.file_based_lock_internal import \
    FileBasedLockInternal


def get_file_based_lock_impl_for_windows() -> Type[FileBasedLockInternal]:
    import ctypes
    import ctypes.wintypes
    import msvcrt
    from ctypes.wintypes import BOOL, DWORD, HANDLE

    # https://learn.microsoft.com/en-us/windows/win32/debug/system-error-codes--0-499-
    W32_ERROR_LOCK_VIOLATION = 33

    kernel32 = ctypes.WinDLL('kernel32', use_last_error=True)

    # https://learn.microsoft.com/en-us/windows/win32/api/fileapi/nf-fileapi-lockfile
    win_lock_file_fn_ref = kernel32.LockFile
    win_lock_file_fn_ref.restype = BOOL
    win_lock_file_fn_ref.argtypes = [HANDLE, DWORD, DWORD, DWORD, DWORD]

    def win_lock_file(hFile, dwFileOffsetLow, dwFileOffsetHigh, nNumberOfBytesToLockLow, nNumberOfBytesToLockHigh) -> bool:
        return win_lock_file_fn_ref(hFile, dwFileOffsetLow, dwFileOffsetHigh, nNumberOfBytesToLockLow, nNumberOfBytesToLockHigh)

    # https://learn.microsoft.com/en-us/windows/win32/api/fileapi/nf-fileapi-unlockfile
    win_unlock_file_fn_ref = kernel32.LockFile
    win_unlock_file_fn_ref.restype = BOOL
    win_unlock_file_fn_ref.argtypes = [HANDLE, DWORD, DWORD, DWORD, DWORD]

    def win_unlock_file(hFile, dwFileOffsetLow, dwFileOffsetHigh, nNumberOfBytesToLockLow, nNumberOfBytesToLockHigh) -> bool:
        return win_unlock_file_fn_ref(hFile, dwFileOffsetLow, dwFileOffsetHigh, nNumberOfBytesToLockLow, nNumberOfBytesToLockHigh)

    class _FileBasedLockWindows(FileBasedLockInternal):

        def __init__(self, lock_file_path):
            super().__init__(lock_file_path)
            self._hFile = None
            self._lock_fd = None
        
        def try_lock(self):
            if self._lock_fd is not None:
                raise Exception('Cannot lock twice!')

            _lock_fd = open(self._lock_file_path, 'a+')
            try:
                _hFile = msvcrt.get_osfhandle(_lock_fd.fileno())
                lock_acquired = win_lock_file(_hFile, 0, 0, 1, 0)
                if lock_acquired:
                    self._lock_fd = _lock_fd
                    self._hFile = _hFile
                    _lock_fd.seek(0)
                    _lock_fd.write(self._get_lock_info())
                    _lock_fd.truncate()
                    _lock_fd.flush()
                    return True


                lock_error = ctypes.get_last_error()

                if lock_error == W32_ERROR_LOCK_VIOLATION:
                    return False
                else:
                    raise Exception(f'Got an unexpected error while attempting to lock: {lock_error}')

            finally:
                if self._lock_fd == None:
                    # failed to acquire lock, close the file
                    _lock_fd.close()

        def release_lock(self):
            self._assert_locked()

            # clear lock file
            self._lock_fd.seek(0)
            self._lock_fd.truncate()
            self._lock_fd.close()

            # remove lock file
            os.unlink(self._lock_file_path)

            # release lock
            win_unlock_file(self._hFile, 0, 0, 1, 0)

            self._hFile = None
            self._lock_fd = None
        
        def _assert_locked(self):
            if self._hFile is None:
                raise Exception('Illegal state: not locked')


    return _FileBasedLockWindows



import os
from glob import glob
from typing import AnyStr

from pyshrimp.exception import IllegalArgumentException


def _ls(list_supplier, dirs_only=False, files_only=False):
    if dirs_only and files_only:
        raise IllegalArgumentException('Illegal arguments - dirs_only and files_only cannot be true together')

    full_paths = [
        # remove trailing separator, but don't do that if the whole string is trailing separator
        # just in case someone actually uses /** which should return the /
        path.rstrip(os.path.sep) if path != os.path.sep else path for path in list_supplier()
    ]

    if dirs_only or files_only:
        return list(
            filter(
                os.path.isdir if dirs_only else os.path.isfile,
                full_paths
            )
        )
    else:
        # no filtering required
        return full_paths


def ls(*path_segments: str, dirs_only=False, files_only=False):
    path = os.path.join(*path_segments)
    return _ls(
        lambda: [
            os.path.join(path, file_name) for file_name in os.listdir(path)
        ],
        dirs_only=dirs_only,
        files_only=files_only
    )


def glob_ls(*path_glob_segments: str, dirs_only=False, files_only=False, recursive=False):
    return _ls(
        lambda: glob(os.path.join(*path_glob_segments), recursive=recursive),
        dirs_only=dirs_only,
        files_only=files_only
    )


def write_to_file(file_path: str, data: AnyStr, open_mode='w', encoding='utf-8') -> None:
    effective_encoding = encoding if 'b' not in open_mode else None
    with open(file_path, open_mode, encoding=effective_encoding) as f:
        f.write(data)


def read_file(file_path: str, encoding='utf-8') -> str:
    """
    Read contents of file using text mode.

    :param file_path:
    :return:
    """
    with open(file_path, 'r', encoding=encoding) as f:
        return f.read()


def read_file_bin(file_path: str) -> bytes:
    """
    Read contents of file using binary mode.

    :param file_path:
    :return:
    """
    with open(file_path, 'rb') as f:
        return f.read()


def chmod_set(file_path, mode_to_set) -> None:
    """
    Updates file mode with given flags (new_mode = current_mode | mode_to_set)

    :param file_path: Path of the target file
    :param mode_to_set: Flags to set, see stat.S_* for values
    :return:
    """
    st = os.stat(file_path)
    os.chmod(file_path, st.st_mode | mode_to_set)


def chmod_unset(file_path, mode_to_remove) -> None:
    """
    Updates file mode by removing given flags (`new_mode = current_mode ^ mode_to_remove`)

    :param file_path: Path of the target file
    :param mode_to_remove: Flags to remove, see stat.S_I* for values
    :return:
    """
    st = os.stat(file_path)
    os.chmod(file_path, st.st_mode ^ mode_to_remove)



import re
from typing import Union, List

from pyshrimp.utils.splitter import Splitter, default_splitter
from pyshrimp.utils.table_parser import parse_table


class StringWrapper(str):

    def lines(self, include_empty=False):
        all_lines = self.splitlines()
        if include_empty:
            return all_lines
        else:
            return [line for line in all_lines if line]

    def match_lines(self, pattern, capture_group: Union[str, int] = 1, include_empty_lines=False) -> List[Union[str, None]]:
        match_list = (
            re.match(pattern, el) for el in self.lines(include_empty_lines)
        )
        return [
            m.group(capture_group) for m in match_list if m
        ]

    def match_lines_multi_group(self, pattern, capture_groups: List[Union[str, int]], include_empty_lines=False) -> List[List[Union[str, None]]]:
        match_list = (
            re.match(pattern, el) for el in self.lines(include_empty_lines)
        )
        return [
            [
                m.group(g) for g in capture_groups
            ] for m in match_list if m
        ]

    def columns(self, *column_index, splitter: Splitter = default_splitter, maxsplit=0):
        def _process_line(line):
            split_line = splitter(line, maxsplit=maxsplit)
            if not column_index:
                return split_line
            else:
                return [(split_line[i:i + 1] or [None])[0] for i in column_index]

        return [
            _process_line(line) for line in self.lines(include_empty=False)
        ]

    def parse_table(self, splitter: Splitter = default_splitter):
        return parse_table(self.lines(include_empty=False), splitter)



import os
from typing import Iterable, List, Optional, TextIO, Union, Dict

# noinspection PyProtectedMember
from pyshrimp._internal.utils.subprocess_utils import _ExecutingProcess, _spawn_process
from pyshrimp.execution_pipeline.pipeline_api import PipelineElement, PipelineExecutionResult
from pyshrimp.utils.collections import first_not_null
from pyshrimp.utils.string_wrapper import StringWrapper
from pyshrimp.utils.subprocess_utils import (
    ProcessExecutionResult,
    run_process
)


class SkipConfig:

    def __init__(
        self,
        skip=False,
        skip_when_no_args=False,
        skipped_out='Execution skipped',
        skipped_err='',
        skipped_code=0
    ):
        self._skip = skip
        self._skip_when_no_args = skip_when_no_args
        self.skipped_code = skipped_code
        self.skipped_err = skipped_err
        self.skipped_out = skipped_out

    def should_skip(self, args):
        if self._skip:
            return True

        if self._skip_when_no_args:
            if not any(args):
                return True

        return False


class CommandArgProcessor:
    def process_args(self, *args):
        raise NotImplementedError()


class DefaultCommandArgProcessor(CommandArgProcessor):

    def process_args(self, *args):
        res = []
        for el in args:
            if isinstance(el, str):
                res.append(el)
            elif isinstance(el, Iterable):
                res += [str(sub_el) for sub_el in el]
            else:
                res.append(str(el))

        return res


class _CommandPipelineElement(PipelineElement):

    def __init__(self, command: 'Command', executing_command: _ExecutingProcess) -> None:
        super().__init__()
        self._executing_command = executing_command
        self._command = command
        self._result: Optional[ProcessExecutionResult] = None
        self._stdout = self._executing_command.stdout

    def _close_once(self):
        self._result = self._executing_command.close()
        self._stdout.close()

    def stdout_for_pipe(self):
        self._executing_command.detach_stdout()
        return self._stdout

    @property
    def result(self) -> PipelineExecutionResult:
        return PipelineExecutionResult(
            stdout=self._result.standard_output,
            stderr=self._result.error_output,
            result=self._result,
            exception=self._result.exception
        )


class Command:
    def __init__(
        self,
        command: List[str],
        check: bool = True,
        capture: bool = True,
        argument_processor: CommandArgProcessor = None,
        env: Dict[str, str] = None,
        env_append: bool = True,
        cwd: Optional[str] = None
    ):
        self._env = env
        self._env_append = env_append
        self._check = check
        self._capture = capture
        self._argument_processor = argument_processor or DefaultCommandArgProcessor()
        # by-design we do not use argument processor this list - just cast everything to string to ensure we won't blow up
        self._command = [str(el) for el in command]
        self._cwd = cwd

    def __call__(self, *args, **kwargs):
        return self.exec(*args, **kwargs)

    def exec(
        self, *args, cmd_in=None,
        check: Optional[bool] = None,
        capture: Optional[bool] = None,
        skip: Union[bool, SkipConfig] = False,
        cwd: Optional[str] = None
    ) -> ProcessExecutionResult:

        command = self._build_command(args)

        if isinstance(skip, bool):
            skip = SkipConfig(skip=skip)

        if skip.should_skip(args):
            return ProcessExecutionResult(
                command=command,
                standard_output=StringWrapper(skip.skipped_out),
                error_output=StringWrapper(skip.skipped_err),
                return_code=skip.skipped_code
            )

        _capture = first_not_null(capture, self._capture)

        res = run_process(
            command=command,
            capture_out=_capture,
            capture_err=_capture,
            cmd_in=cmd_in,
            env=self._build_env(),
            cwd=first_not_null(cwd, self._cwd)
        )

        if first_not_null(check, self._check):
            res.raise_if_not_ok()

        return res

    def _build_env(self):
        env = os.environ.copy() if self._env_append else {}
        env.update(self._env or {})
        return env

    def with_args(self, *args):
        return Command(
            command=self._build_command(args=args),
            check=self._check,
            capture=self._capture,
            argument_processor=self._argument_processor,
            env=self._env,
            env_append=self._env_append,
            cwd=self._cwd,
        )

    def __str__(self):
        return f'Command({self.__dict__})'

    def _build_command(self, args=None):
        command_args = self._argument_processor.process_args(*(args or []))
        return self._command + command_args

    def pipe_connect_async(self, left_out: TextIO) -> _CommandPipelineElement:
        return _CommandPipelineElement(
            command=self,
            executing_command=_spawn_process(
                command=self._build_command(),
                cmd_in=left_out,
                env=self._build_env(),
                capture_output=self._capture,
                capture_err_output=self._capture,
                cwd=self._cwd
                # TODO: other params like check
            )
        )

    def _spawn(self):
        return _spawn_process(
            command=self._build_command(),
            env=self._build_env(),
            capture_output=self._capture,
            capture_err_output=self._capture,
            cwd=self._cwd
            # TODO: other params like check
        )


def cmd(command: Union[str, List], *args, check=True, capture=True, env: Dict[str, str] = None, env_append=True, cwd: str = None) -> Command:
    if isinstance(command, str):
        command = [command]

    return Command(
        command=command + list(args),
        check=check,
        capture=capture,
        env=env,
        env_append=env_append,
        cwd=cwd
    )


def shell_cmd(script: str, *args: str, check=True, capture=True, env: Dict[str, str] = None, env_append=True, cwd: str = None) -> Command:
    return Command(
        command=['/bin/bash', '-c', script, 'bash'] + [str(el) for el in args],
        check=check,
        capture=capture,
        env=env,
        env_append=env_append,
        cwd=cwd
    )



import inspect
import os
import sys
from threading import Thread
from typing import Callable, List

# noinspection PyProtectedMember
from pyshrimp._internal.pipes.async_function import _AsyncFunctionPipelineElement
from pyshrimp.exception import IllegalStateException, IllegalArgumentException
from pyshrimp.execution_pipeline.pipeline_api import PipelineElement, StringPipelineElement, StreamPipelineElement, PipelineExecutionResult, PipelineTerminator, PipelineTerminatorStdout
from pyshrimp.utils.command import cmd, shell_cmd


class ExecutionPipeline:

    def __init__(self) -> None:
        self._items: List[PipelineElement] = []

    def attach_stdin(self):
        if self._items:
            raise IllegalStateException('Attaching STDIN makes only sense at the start of pipeline.')

        self._items.append(StreamPipelineElement(sys.stdin))
        return self

    def attach_text(self, text: str):
        if self._items:
            raise IllegalStateException('Attaching text makes only sense at the start of pipeline.')

        self._items.append(StringPipelineElement(text))
        return self

    def attach_function(self, fn):
        fn_args = dict(inspect.signature(fn).parameters.items())

        if 'stream_input' in fn_args and 'stream_output' in fn_args:
            # async function
            self._attach_async_connector(
                connect_method=lambda left_out: _AsyncFunctionPipelineElement(function=fn, left_out=left_out)
            )

        elif 'stream_input' in fn_args or 'stream_output' in fn_args:
            # invalid async function
            raise IllegalArgumentException(
                f'Asynchronous function must accept both stream_input and stream_output but found only one.'
                f' Function args: {", ".join(fn_args.keys())}'
            )

        else:
            # sync function
            self._attach_sync_connector(
                connect_method=lambda left_out: StringPipelineElement(fn(left_out))
            )

        return self

    def _get_left(self):
        return self._items[-1] if self._items else None

    def _attach_async_connector(self, connect_method):
        left = self._get_left()
        left_out = left.stdout_for_pipe() if left else None

        if isinstance(left_out, str):
            # Connector expects to see streaming input, transform the string to TextIO
            pipe_r, pipe_w = os.pipe()

            input_for_pipe = left_out

            def _write_to_pipe():
                os.write(pipe_w, input_for_pipe.encode('utf-8'))
                os.close(pipe_w)

            # write to pipe in new thread
            # this should workaround any deadlock situations
            Thread(target=_write_to_pipe).start()

            left_out = pipe_r

        self._items.append(connect_method(left_out))

    def _attach_sync_connector(self, connect_method: Callable[[str], PipelineElement]):
        left = self._get_left()

        # We need to consume output as the connector does not support streaming
        if left:
            left.close()
            left_out = left.result.stdout
        else:
            left_out = None

        self._items.append(connect_method(left_out))

    def attach(self, right):
        if isinstance(right, PipelineTerminator):
            return self.close()

        elif isinstance(right, PipelineTerminatorStdout):
            return self.close().stdout

        elif isinstance(right, str):
            self.attach(
                shell_cmd(
                    script=right,
                    check=False
                )
            )

        elif isinstance(right, list):
            self.attach(
                cmd(
                    command=right,
                    check=False
                )
            )

        elif hasattr(right, 'pipe_connect_async'):
            self._attach_async_connector(right.pipe_connect_async)

        elif hasattr(right, 'pipe_connect_sync'):
            self._attach_sync_connector(right.pipe_connect_sync)

        elif callable(right):
            self.attach_function(right)

        else:
            raise Exception(f'Unsupported type for right: {type(right)}')

        return self

    def attach_all(self, *elements):
        for el in elements:
            self.attach(el)

        return self

    def close(self) -> PipelineExecutionResult:
        for el in self._items:
            el.close()

        return self._get_left().result

    def __or__(self, other) -> 'ExecutionPipeline':
        return self.attach(other)

    def __ror__(self, other):
        raise Exception(
            f'Unsupported pipe - left side is not a pipe! Tried to connect the {other} with {self}'
        )


def pipe(*elements) -> PipelineExecutionResult:
    return ExecutionPipeline().attach_all(*elements).close()



from dataclasses import dataclass
from typing import Any, List, Tuple, Union

_dot_dict_state_field_name = '_DotDict__dot_dict_state_d40de1e455ae'


def _format_path(path):
    res = []

    for idx, (el_name, sep) in enumerate(path):
        if idx > 0:
            res.append(sep)

        res.append(str(el_name))

    return ''.join(res)


@dataclass
class _DotDictState:
    data: Any
    path: List[Tuple[str, str]]


def _map_item(dot_dict, p, v):
    if isinstance(v, dict):
        return DotDict(
            data=v,
            path=getattr(dot_dict, _dot_dict_state_field_name).path + p
        )

    elif isinstance(v, list):
        return [_map_item(dot_dict, p + [(f'[{idx}]', '')], el) for idx, el in enumerate(v)]

    else:
        return v


class DotDict:
    """
    Dict wrapper class designed to enable property-like access to dictionary elements.

    Example:

            >>> d = DotDict(original_dict)
            >>> print(d.a.b.c)

    """

    def __init__(self, data, path=None):
        # use some weird name that have very low chance of colliding with actual key in target dict
        # the full version of this name must match the _dot_dict_state_field_name
        self.__dot_dict_state_d40de1e455ae = _DotDictState(
            data=data,
            path=path or []
        )

    def __contains__(self, item):
        return item in self.__dot_dict_state_d40de1e455ae.data

    def __getattr__(self, item):
        try:
            return _map_item(self, [(item, '.')], self.__dot_dict_state_d40de1e455ae.data[item])
        except KeyError:
            raise KeyError(_format_path(self.__dot_dict_state_d40de1e455ae.path + [(item, '.')]))

    def __setattr__(self, key, value):
        if key == _dot_dict_state_field_name:
            super().__setattr__(key, value)
        else:
            self.__dot_dict_state_d40de1e455ae.data[key] = value

    def __repr__(self):
        return repr(self.__dot_dict_state_d40de1e455ae.data)

    def __str__(self):
        return str(self.__dot_dict_state_d40de1e455ae.data)

    def __getitem__(self, item):
        return self.__getattr__(item)

    def __setitem__(self, key, value):
        self.__setattr__(key, value)

    def __len__(self):
        return len(self.__dot_dict_state_d40de1e455ae.data)

    def __eq__(self, other):
        """
        Compare DotDict instance with other object.
        When two DotDicts are compared their data is compared.
        When dot dict is compared with other type the data is compared with the other type.
        The path is ignored by design.
        """
        if isinstance(other, DotDict):
            return self.__dot_dict_state_d40de1e455ae.data == other.__dot_dict_state_d40de1e455ae.data
        else:
            return self.__dot_dict_state_d40de1e455ae.data == other


def as_dot_dict(dict_data, variable_name='dict'):
    """
    Creates new DotDict instance
    :param dict_data:
    :param variable_name:
    :return:
    """
    return DotDict(dict_data, [(variable_name, '.')])


def unwrap_dot_dict(wrapped_data: Union[DotDict, List]) -> Union[dict, List]:
    if isinstance(wrapped_data, DotDict):
        # noinspection PyProtectedMember
        return wrapped_data._DotDict__dot_dict_state_d40de1e455ae.data

    elif isinstance(wrapped_data, list):
        return [unwrap_dot_dict(el) for el in wrapped_data]

    else:
        return wrapped_data



import os.path
import stat
import sys
import textwrap

# noinspection PyProtectedMember
from pyshrimp._internal.utils.errors import _exit_error


def _dedent(content: str):
    return textwrap.dedent(
        content.strip('\n')
    )


def print_multiline(content: str):
    print(_dedent(content))


# noinspection PyUnusedLocal
def _cmd_help(cmd, args):
    print_multiline(
        '''
        The pyshrimp is wrapper tool for python scripts.
        
        Usages:
        
        1) pyshrimp [target-script.py] [arg]...
        
        Executes target script with PyShrimp. This is also used in case of she-bang in script.
        Arguments will be passed to the target script as standard arguments.
        
        2) pyshrimp [command] [arg]...
        
        Executes CLI command.
        
        COMMANDS:
        
            [help|--help]  
                Prints this message
                
            [generate|new] [new-script-name.py]
                Creates new script.
                If the new script name is not provided prints the script to standard output.  
                
        ENVIRONMENT VARIABLES:
        
            PYSHRIMP_LOG - if set to 1 the bootstrap process will provide additional diagnostic output
            PYSHRIMP_CACHE_DIR - location of cache directory for PyShrimp (default: ~/.cache/pyshrimp) 
            
        '''
    )


def _cmd_generate(cmd, args):
    if len(args) > 1:
        raise _exit_error(f'Too many arguments provided for `{cmd}`: {" ".join(args)}')

    script_content = _dedent(
        '''
        #!/usr/bin/env pyshrimp
        # $requires: 
        
        from pyshrimp import run, log
        
        
        def main():
            log('Hello world, this script uses devloop!')
        
            
        run(main, devloop=True)
        '''
    )

    if args:
        new_script_path = os.path.abspath(args[0])
        if os.path.exists(new_script_path):
            raise _exit_error(
                f'Target script exists, refusing to override. '
                f'Please provide non existing file name. '
                f'Provided name resolved to: {new_script_path}'
            )

        print(f'Writing content to new script: {new_script_path}')

        with open(new_script_path, 'w') as f:
            f.write(script_content)

        # make the script writable
        st = os.stat(new_script_path)
        os.chmod(new_script_path, st.st_mode | stat.S_IXUSR)

    else:
        print(script_content)


_commands = {
    '--help': _cmd_help,
    'help': _cmd_help,
    '-help': _cmd_help,
    'generate': _cmd_generate,
    'new': _cmd_generate
}


def _handle_cli_maybe() -> bool:
    if len(sys.argv) > 1:
        command = sys.argv[1]
        args = sys.argv[2:]

        handler = _commands.get(command)
        if handler:
            handler(command, args)
            return True

    return False



import subprocess
from typing import Union, Iterable

from pyshrimp.utils.string_wrapper import StringWrapper
from pyshrimp.utils.subprocess_utils import ProcessExecutionResult


class _ExecutingProcess:

    def __init__(self, command: Union[str, Iterable], process: subprocess.Popen) -> None:
        self._process = process
        self._command = command
        self._closed = False
        self._result = None

    def close(self, timeout=None) -> ProcessExecutionResult:
        # TODO: try catch

        if not self._closed:
            (out, err) = self._process.communicate(timeout=timeout)
            self._result = ProcessExecutionResult(
                self._command, StringWrapper(out), StringWrapper(err), self._process.returncode
            )
            self._closed = True

        return self._result

    @property
    def stdout(self):
        return self._process.stdout

    def detach_stdout(self):
        self._process.stdout = None

    def send_signal(self, signal):
        self._process.send_signal(signal)

    @property
    def process(self) -> subprocess.Popen:
        return self._process

    @property
    def pid(self) -> int:
        return self._process.pid


def _spawn_process(command: Union[str, Iterable], cmd_in=None, cwd=None, env=None, capture_output=False, capture_err_output=False) -> _ExecutingProcess:
    # TODO: support for string stdin?
    # TODO: try catch support
    process = subprocess.Popen(
        command,
        stdin=subprocess.DEVNULL if cmd_in is None else cmd_in,
        stdout=subprocess.PIPE if capture_output else None,
        stderr=subprocess.PIPE if capture_err_output else None,
        # TODO: binary output support
        universal_newlines=True,
        shell=False,
        cwd=cwd,
        env=env
    )

    return _ExecutingProcess(command, process)



from dataclasses import dataclass
from typing import List

from pyshrimp.utils.dotdict import DotDict
from pyshrimp.utils.splitter import Splitter, default_splitter


@dataclass
class ParsedTable:
    header: List[str]
    rows: List[List[str]]

    def dict_rows(self, use_dot_dict=True):
        for row in self.rows:
            row_dict = dict(zip(self.header, row))
            yield DotDict(row_dict) if use_dot_dict else row_dict


def parse_table(
    lines: List[str],
    splitter: Splitter = default_splitter,
) -> ParsedTable:
    header = None
    maxsplit = 0
    rows = []
    for line in lines:
        if not header:
            # parse header
            header = splitter(line)
            maxsplit = len(header) - 1
        else:
            # parse row
            if maxsplit == 0:
                # special case: single column == whole line should be used (no split at all)
                rows.append([line])
            else:
                rows.append(splitter(line, maxsplit=maxsplit))

    return ParsedTable(header, rows)
